rawset(_G, "kartzshields", {})
kartzshields.debug = false

if (kartzshields.debug)
	kartzshields.cv_x = CV_RegisterVar({
		name = "hudx",
		defaultvalue = 0,
		PossibleValue = {MIN = -999999999, MAX = 999999999},	-- CV_Unsigned but has lower bound maximized
		flags = CV_NETVAR,
	})
	kartzshields.cv_y = CV_RegisterVar({
		name = "hudy",
		defaultvalue = 0,
		PossibleValue = {MIN = -999999999, MAX = 999999999},
		flags = CV_NETVAR,
	})
end

-- sounds
for i = 1, 8
	freeslot("sfx_shrol"..i)
end

-- function to get player count
kartzshields.GetPlayerCount = function()
	local pcount = 0
	
	for p in players.iterate
		if (p.spectator) then continue end
		pcount = $+1
	end
	
	return pcount
end

-- return split flags
kartzshields.ReturnSplitFlags = function(f, displaynum)
	if not (splitscreen) return f end

	-- exhibit 1: 2p
	if (splitscreen == 1)
		if not (displaynum) -- p1
			return f & ~(V_SNAPTOBOTTOM)
		else	-- p2
			return (f|V_SPLITSCREEN) & ~(V_SNAPTOTOP)
		end
	else
		-- jesus fuck this sucks dicks
		if not (displaynum)	-- p1
			return f & ~(V_SNAPTOBOTTOM|V_SNAPTORIGHT)
		elseif (displaynum == 1)	-- p2
			return (f|V_HORZSCREEN) & ~(V_SNAPTOBOTTOM|V_SNAPTOLEFT)
		elseif (displaynum == 2)	-- p3
			return (f|V_SPLITSCREEN) & ~(V_SNAPTOTOP|V_SNAPTORIGHT)
		elseif (displaynum == 3)	-- p4
			return (f|V_SPLITSCREEN|V_HORZSCREEN) & ~(V_SNAPTOTOP|V_SNAPTOLEFT)
		end
	end
	return f	-- shouldn't happen but don't return nil either way lol
end

kartzshields.IsDisplayPlayer = function(p)
	for i = 0, (splitscreen)	-- displayplayers.iterate breaks replays rn so we'll work with this
		local dp = displayplayers[i]

		if (p == dp) then return i end	-- return their index if they're a displayplayer too
	end
	-- if we're here, we're not a displayplayer
	return false
end

-- dummy shield object
freeslot("MT_KZS_SHIELD")
mobjinfo[MT_KZS_SHIELD] = {
	spawnstate = S_INVISIBLE,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY|MF_DONTENCOREMAP,
}
addHook("MobjThinker", function(shield)
	local target = shield.target
	if not (target and target.valid) then P_RemoveMobj(shield) return end
	
	local owner = target.player
	if not (owner.kzs and owner.kzs.shield) then P_RemoveMobj(shield) return end
	if (owner.kzs.shield != shield.kzs_shield) then P_RemoveMobj(shield) return end
end, MT_KZS_SHIELD)

-- fake item box (for the attraction shield)
freeslot("MT_KZS_ITEMBOX")
mobjinfo[MT_KZS_ITEMBOX] = {
	spawnstate = S_RANDOMITEM1,
	radius = 48*FRACUNIT,
	height = 48*FRACUNIT,
	flags = MF_SLIDEME|MF_SPECIAL|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_DONTENCOREMAP,
}
addHook("MobjThinker", function(box)
	if not (box.target and box.target.valid)
	or not (box.target.player)
		-- spawn a pop object
		local pop = P_SpawnMobj(box.x, box.y, box.z, MT_THOK)
		pop.state = S_XPLD1
		if (box.eflags & MFE_VERTICALFLIP)
			pop.eflags = $|(MFE_VERTICALFLIP)
			pop.flags2 = $|(MF2_OBJECTFLIP)
		end
		pop.scale = box.scale
		pop.destscale = box.destscale
		S_StartSound(pop, sfx_kc2e)
		
		-- remove the tracer
		P_RemoveMobj(box)
	end
	
	-- follow target
	box.color = box.target.player.skincolor
	box.kzs_chasetimer = $ and $+1 or 1
	A_HomingChase(box, (FRACUNIT*6) * box.kzs_chasetimer)
	if (box.kzs_chasetimer % 2)
		local g = P_SpawnGhostMobj(box)
		g.tics = 3
	end
end, MT_KZS_ITEMBOX)
addHook("TouchSpecial", function(box, mo)
	if (box.target != mo) then return true end
	
	-- spawn a pop object
	local pop = P_SpawnMobj(box.x, box.y, box.z, MT_THOK)
	pop.state = S_XPLD1
	if (box.eflags & MFE_VERTICALFLIP)
		pop.eflags = $|(MFE_VERTICALFLIP)
		pop.flags2 = $|(MF2_OBJECTFLIP)
	end
	pop.scale = box.scale
	pop.destscale = box.destscale
	S_StartSound(pop, sfx_kc2e)
	
	-- start the roulette
	if not (mo.player.kartstuff[k_itemroulette])
	and not (mo.player.kartstuff[k_itemtype])
		mo.player.kartstuff[k_itemroulette] = 1
	end
	
	-- remove the tracer
	P_RemoveMobj(box)
end, MT_KZS_ITEMBOX)

-- fire (for the elemental shield)
freeslot("S_KZS_ELEMENTALFIRE", "MT_KZS_ELEMENTALFIRE", "SPR_ESFL")
states[S_KZS_ELEMENTALFIRE] = {SPR_ESFL, A|FF_ANIMATE|FF_FULLBRIGHT, TICRATE*3, nil, 5, 2, S_NULL}
mobjinfo[MT_KZS_ELEMENTALFIRE] = {
	spawnstate = S_KZS_ELEMENTALFIRE,
	deathstate = S_NULL,
	radius = 8*FRACUNIT,
	height = 14*FRACUNIT,
	flags = MF_DONTENCOREMAP,
}
addHook("MobjThinker", function(fire)
	local colour = SKINCOLOR_KETCHUP	-- default colour
	if (fire.target and fire.target.player)
		colour = fire.target.player.skincolor
	end
	fire.color = colour
end, MT_KZS_ELEMENTALFIRE)
local function elementalCollide(fire, mo)
	if (fire.type != MT_KZS_ELEMENTALFIRE) then return end	-- weird paranoia

	if not (mo.health) then return end	-- don't damage dead things
	if not (mo.flags & MF_SHOOTABLE) then return end	-- damage shootables!
	if ((fire.target and fire.target.valid) and (fire.target == mo)) then return end	-- don't damage the spawner!
	
	if not (abs(fire.z - mo.z) <= fire.height) then return end	-- height check
	
	if (mo.player)
		local p = mo.player
		if (p.powers[pw_flashing])
		or (p.kartstuff[k_invincibilitytimer])
		or (p.kartstuff[k_growshrinktimer] > 0)
			return 
		end
		
		-- spin them out!
		K_SpinPlayer(p, fire.target, 0, fire)
		
		-- hitfeed support
		if (hitfeed)
		and ((fire.target and fire.target.valid) and fire.target.player)
			local source = fire.target
			
			local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
			local v = {{p.name, p.mo.hf_overridenamecolor or p.skincolor}, {skins[p.mo.skin].facemmap, p.skincolor}, p}
			if (s[3] == v[3]) then s = nil end	-- source == victim, remove source
			
			HF_SendHitMessage(s, v, "K_HMELEM")
		end
	else
		local clashItem = false
		
		local clashItems = {
			MT_JAWZ_DUD, MT_JAWZ_SHIELD,
			MT_ORBINAUT, MT_ORBINAUT_SHIELD,
			MT_BANANA, MT_BANANA_SHIELD,
		}
		for _, item in ipairs(clashItems)
			if (mo.type != item) then continue end
			
			clashItem = true
			break
		end
		
		if (clashItem)	-- these need more effects
			-- noclipthing the item
			mo.flags = $|MF_NOCLIPTHING
			P_SetOrigin(mo, mo.x, mo.y, mo.z + (mo.height * P_MobjFlip(mo)))
			S_StartSound(mo, mo.info.deathsound)
			
			-- kill and launch
			P_KillMobj(mo)
			P_SetObjectMomZ(mo, FRACUNIT*8)
			P_InstaThrust(mo, P_RandomRange(0, 359)*ANG1, 16*FRACUNIT)
			
			-- spawn clash effect
			P_SpawnMobj(mo.x, mo.y, mo.z, MT_ITEMCLASH)
		else
			P_DamageMobj(mo, fire, fire.target, 1)
		end
	end
end
addHook("MobjCollide", elementalCollide, MT_KZS_ELEMENTALFIRE)
addHook("MobjMoveCollide", elementalCollide, MT_KZS_ELEMENTALFIRE)