-- main thinkframe
addHook("ThinkFrame", do
	-- hitfeed support
	if not (G_BattleGametype())
	and (hitfeed)
		HF_OverrideLocation(76)
	end
	
	-- player thinkers
	for p in players.iterate
		-- create namespace
		p.kzs = $ or {}
		p.kzs_shieldbutton = (p.cmd.buttons & BT_CUSTOM1) and ($ and $+1 or 1) or 0	-- this is separate from the framework
		
		-- functions
		kartzshields.LapRoulette(p)	-- roll shields on lap completion
		kartzshields.Roulette(p)	-- roll the roulette for shields
		kartzshields.ShieldThinker(p)	-- main thinker for shields
	end
end)

-- damage stuff
addHook("PlayerSpin", function(p, inflictor, source)
	if (p.powers[pw_flashing]) then return end
	if (p.pflags & PF_TIMEOVER) then return end	-- LMFAO
	local mo = p.mo
	
	if not (kartzshields.HasShield(p)) then return end
	if not (p.kzs.usingshield) then return end
	local info = kartzshields.Shields[p.kzs.shield]
	
	if (info.playerspin)
		local playerspin = info.playerspin(p, inflictor, source)
		if (playerspin) then return playerspin end	-- override
	end
	
	-- remove health
	p.kzs.health = $-1
	if (p.kzs.health)
		p.powers[pw_flashing] = TICRATE*5/2
		S_StartSound(mo, sfx_shield)
		return true
	else
		-- remove shield
		mo.kzs_shieldhack = true	-- triggers a hack
		kartzshields.RemoveShield(p)
		S_StartSound(mo, sfx_shldls)
	end
end)
addHook("PlayerSquish", function(p, inflictor, source)
	if (p.powers[pw_flashing]) then return end
	if (p.pflags & PF_TIMEOVER) then return end	-- LMFAO
	local mo = p.mo
	
	if not (kartzshields.HasShield(p)) then return end
	if not (p.kzs.usingshield) then return end
	local info = kartzshields.Shields[p.kzs.shield]
	
	if (info.playersquish)
		local playersquish = info.playersquish(p, inflictor, source)
		if (playersquish) then return playersquish end	-- override
	end
	
	-- remove health
	p.kzs.health = $-1
	if (p.kzs.health)
		p.powers[pw_flashing] = TICRATE*5/2
		S_StartSound(mo, sfx_shield)
		return true
	else
		-- remove shield
		mo.kzs_shieldhack = true	-- triggers a hack
		kartzshields.RemoveShield(p)
		S_StartSound(mo, sfx_shldls)
	end
end)
addHook("PlayerExplode", function(p, inflictor, source)
	if (p.powers[pw_flashing]) then return end
	if (p.pflags & PF_TIMEOVER) then return end	-- LMFAO
	local mo = p.mo
	
	if not (kartzshields.HasShield(p)) then return end
	if not (p.kzs.usingshield)
		kartzshields.RemoveShield(p)
		return 
	end
	local info = kartzshields.Shields[p.kzs.shield]
	
	if (info.playerexplode)
		local playerexplode = info.playerexplode(p, inflictor, source)
		if (playerexplode) then return playerexplode end	-- override
	end
	
	-- remove health
	p.kzs.health = $-1
	if (p.kzs.health)
		p.powers[pw_flashing] = TICRATE*5/2
		S_StartSound(mo, sfx_shield)
		return true
	else
		-- remove shield
		mo.kzs_shieldhack = true	-- triggers a hack
		kartzshields.RemoveShield(p)
		S_StartSound(mo, sfx_shldls)
	end
end)
addHook("MobjDamage", function(mo, inflictor, source)
	if ((inflictor and inflictor.valid) and inflictor.type == MT_SINK) then return end
	local p = mo.player
	if (p.powers[pw_flashing]) then return end
	if (p.pflags & PF_TIMEOVER) then return end	-- LMFAO
	
	-- gotta check these for mobjdamage...
	if (p.kartstuff[k_invincibilitytimer])
	or (p.kartstuff[k_growshrinktimer] > 0)
	or (p.kartstuff[k_hyudorotimer])
		return
	end
	
	if not (kartzshields.HasShield(p)) then return end
	if not (p.kzs.usingshield) then return end
	local info = kartzshields.Shields[p.kzs.shield]
	
	if (info.mobjdamage)
		local mobjdamage = info.mobjdamage(mo, inflictor, source)
		if (mobjdamage) then return mobjdamage end	-- override
	end
	
	-- remove health
	p.kzs.health = $-1
	
	if (p.kzs.health)
		p.powers[pw_flashing] = TICRATE*5/2
		S_StartSound(mo, sfx_shield)
		return true
	else
		-- remove shield
		mo.kzs_shieldhack = true	-- triggers a hack
		kartzshields.RemoveShield(p)
		S_StartSound(mo, sfx_shldls)
	end
end, MT_PLAYER)

-- render the shield box
hud.add(function(v, p, c)
	if not (p.kzs) then return end
	if not (p.kzs.roulette or p.kzs.shield) then return end
	
	-- positional vars
	local dp = kartzshields.IsDisplayPlayer(p)
	
	local hx = 39
	local hy = (splitscreen) and -5 or -3
	local flags = kartzshields.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_HUDTRANS, dp)
	
	if (splitscreen > 1)
		local right = (dp == 1 or dp == 3)
		
		hx = ((right) and 121 or -9) + (23 * ((right) and -1 or 1))
		hy = -8
		flags = kartzshields.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
	end
	
	-- the backdrop
	v.draw(hx, hy, v.cachePatch("K_SHBG"), flags)
	
	if (p.kzs.roulette)
		local shield = (p.kzs.roulette - 1) % (3 * #kartzshields.Shields) / 3
		local patch = kartzshields.Shields[shield + 1].patch
		
		v.draw(hx, hy, v.cachePatch(patch), flags, v.getColormap(TC_RAINBOW, p.skincolor))
	elseif (p.kzs.shield)
		local patch = kartzshields.Shields[p.kzs.shield].patch
		
		local skin = TC_DEFAULT
		local colour = SKINCOLOR_NONE
		if (p.kzs.shieldblink and (leveltime%2))
			skin = TC_BLINK
			colour = (p.kzs.shieldblinkmode == 2) and SKINCOLOR_RED or SKINCOLOR_WHITE
		end
		
		local flash = (p.kzs.usingshield)
		if not (flash and (leveltime%2))
			v.draw(hx, hy, v.cachePatch(patch), flags, v.getColormap(skin, colour))
		end
	end
end)

-- reset namespaces for players
local function resetFunc()
	-- reset namespaces
	for p in players.iterate
		p.kzs = {}
	end
end
addHook("MapChange", resetFunc)
addHook("MapLoad", resetFunc)	-- sometimes shields wouldnt be reset i guess

-- custom functionality for item boxes
addHook("MobjThinker", function(item)
	if not (item and item.valid) then return end
	
	if (item.kzs_fuse)
		item.flags2 = $|MF2_DONTDRAW
		
		-- make sure they have no target alive preventing fuse from going down
		if (item.kzs_fusetarget and item.kzs_fusetarget.valid) then return end
		
		-- decrease fuse
		item.kzs_fuse = $-1
		if not (item.kzs_fuse)
			item.flags2 = $ & ~(MF2_DONTDRAW)
		end
	end
end, MT_RANDOMITEM)
addHook("MobjMoveCollide", function(mo, item)
	if not (item and item.valid) then return end

	if (item.kzs_fuse)
		return false -- disable colision
	end
end, MT_PLAYER)