-- item odds garbo
kartzshields.ShieldOdds = {
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- pity shield
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- force shield
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- attraction shield
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- elemental shield
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- whirlwind shield
	//{20, 0, 0, 0, 0, 0, 0, 0},	-- armageddon shield
}

if (kartzshields.debug)
	COM_AddCommand("modifyshieldodds", function(p, shield, ...)
		shield = tonumber($)
		if not (shield)
		or not (kartzshields.ShieldOdds[shield])
			return
		end
		
		-- validate if we can set new odds
		local shieldodds = {...}
		local oddsvalid = true
		
		if (#shieldodds != 8)
			oddsvalid = false
		end
		if (oddsvalid)
			for i = 1, #shieldodds
				local odds = shieldodds[i]
				if (tonumber(odds) != nil) then continue end
				oddsvalid = false
				break
			end
		end
		
		-- set new odds
		if (oddsvalid) then kartzshields.ShieldOdds[shield] = shieldodds end
		
		-- print new odds
		local odds_string = ""
		for pos, odds in ipairs(kartzshields.ShieldOdds[shield])
			odds_string = $ + tostring(odds)
			if (pos != 8) then odds_string = $ + ", " end
		end
		CONS_Printf(p, kartzshields.Shields[shield].name.." odds: "..odds_string)
	end, COM_ADMIN)
end

kartzshields.FindUseOdds = function(p, mashed, pingame, spbrush)
	local oddsvalid = {}
	for pos = 1, 8	-- always 8 indexes for odds
		local available = false
		
		for item = 1, #kartzshields.ShieldOdds
			if (kartzshields.GetShieldOdds(pos, item, mashed, pingame) > 0)
				-- if we're here, we have at least one item for grabs
				available = true
				break
			end
		end
		
		oddsvalid[pos] = available
	end
	
	-- get player distance from first to last
	local pdis = 0
	for player in players.iterate
		if (player.spectator or not player.mo) then continue end	-- not a valid player
		if not (player.kartstuff[k_position] != 0) then continue end	-- doesn't have a position?
		if not (player.kartstuff[k_position] < p.kartstuff[k_position]) then continue end	-- not behind the person getting a shield
		
		local xdiff = player.mo.x - p.mo.x
		local ydiff = player.mo.y - p.mo.y
		local zdiff = player.mo.z - p.mo.z
		pdis = $ + FixedHypot(FixedHypot(xdiff, ydiff), zdiff) / mapobjectscale * (pingame - player.kartstuff[k_position]) / max(1, ((pingame - 1) * (pingame + 1) / 3))
	end
	
	-- setup dist table
	local disttable, distlen = {}, 0
	local function kzs_setupdisttable(odds, num)
		if not (oddsvalid[odds]) then return end
		
		for i = num, 1, -1
			disttable[distlen] = odds
			distlen = $+1
		end
	end
	kzs_setupdisttable(1, 1)
	kzs_setupdisttable(2, 1)
	kzs_setupdisttable(3, 1)
	kzs_setupdisttable(4, 2)
	kzs_setupdisttable(5, 2)
	kzs_setupdisttable(6, 3)
	kzs_setupdisttable(7, 3)
	kzs_setupdisttable(8, 1)
	
	-- pdis modifiers
	if (franticitems) then pdis = (15 * $) / 14 end	-- frantic items
	if (spbrush) then pdis = (3 * $) / 2 end	-- spb rush
	pdis = ((28 + (8 - pingame)) * $) / 28	-- ??
	
	-- find useodds based on distance
	local distvar = 64 * 14	-- ARGH
	local useodds = 0
	
	if (pdis <= 0)	-- no distance?
		useodds = disttable[0]
	elseif (pdis > distvar * ((12 * distlen) / 14))	-- yeah
		useodds = disttable[distlen-1]
	else
		for i = 1, 13
			if not (pdis <= distvar * ((i * distlen) / 14)) then continue end
			useodds = disttable[((i * distlen) / 14)]
			break
		end
	end
	
	return useodds
end

kartzshields.GetShieldOdds = function(pos, item, mashed, pingame, spbrush)
	-- TODO: shield disabling?	
	local newodds = kartzshields.ShieldOdds[item][pos] * 4	-- *4 for fractional precision
	
	-- power item modifiers
	local playerscaling = (8 - (spbrush and 2 or pingame))
	for item = 1, #kartzshields.Shields
		if not (kartzshields.Shields[item].power) then continue end
		
		if (franticitems) then newodds = $<<1 end
		newodds = FixedMul($<<FRACBITS, FRACUNIT+((playerscaling<<FRACBITS)/25))>>FRACBITS
		if (mashed > 0) then newodds = FixedDiv($<<FRACBITS, FRACUNIT + mashed)>>FRACBITS end
	end
	
	return newodds
end

kartzshields.RollShieldOdds = function(p, mashed)
	local pingame = kartzshields.GetPlayerCount()
	local spbrush = (spbplace != -1 and p.kartstuff[k_position] == spbplace+1)
	
	-- init spawnchance values
	local spawnchance, totalspawnchance = {}, 0
	for item = 0, #kartzshields.ShieldOdds
		spawnchance[item] = 0
	end
	
	-- find player useodds
	local useodds = kartzshields.FindUseOdds(p, mashed, pingame, spbrush)
	
	-- calculate totalspawnchance
	for item = 1, #kartzshields.ShieldOdds
		totalspawnchance = $ + kartzshields.GetShieldOdds(useodds, item, mashed, pingame, spbrush)
		spawnchance[item] = totalspawnchance
	end
	if not (totalspawnchance) then return end	-- nope
	
	-- get the shield!
	totalspawnchance = P_RandomKey($)
	local shield
	for item = 0, #kartzshields.ShieldOdds
		shield = item
		if (spawnchance[item] > totalspawnchance) then break end	-- this is it
	end
	
	return shield
end