-- easy function to check if we have a shield
kartzshields.HasShield = function(p)
	return (p.kzs and p.kzs.shield)
end

-- adding shields (this has no validation, use with care lol)
kartzshields.Shields = {}

local kzs_shields = {
	-- pity shield
	{
		name = "Pity Shield",
		state = S_PITY1,
		patch = "K_SHLD_1",
		odds = {8, 0, 0, 0, 0, 0, 0, 0},
	},
	
	-- force shield
	{
		name = "Force Shield",
		state = S_FORC1,
		health = 2,
		patch = "K_SHLD_2",
		
		passive = function(p)
			local mo = p.mo
			
			if (p.kzs.health == 1)
			and not (mo.kzs_orb.forceweak)
				mo.kzs_orb.state = S_FORC11
				mo.kzs_orb.forceweak = true
			end
		end,
		
		odds = {0, 0, 1, 2, 3, 2, 0, 0},
	},
	
	-- attraction shield
	{
		name = "Attraction Shield",
		state = S_MAGN1,
		patch = "K_SHLD_3",
		
		activate = function(p)
			local mo = p.mo
			
			p.kzs.attraction = 0
			K_PlayAttackTaunt(mo)
			S_StartSound(mo, sfx_s3k45)
		end,
		active = function(p)
			local mo = p.mo
			
			local range = FixedMul(mo.scale, 384*FRACUNIT)
			local height = FixedMul(FRACUNIT*3/4, range)
			searchBlockmap("objects", function(ref, found)
				if (ref == found) then return end	-- don't infect ourself!
				if (found.type != MT_RANDOMITEM) then return end	-- only affect random items
				if (found.fuse or found.kzs_fuse) then return end	-- they're respawning!
				
				-- positional checks
				if not (P_CheckSight(ref, found)) then return end	-- LoS check
				if (R_PointToDist2(ref.x, ref.y, found.x, found.y) > range) then return end	-- make it radial
				if not (abs(ref.z - found.z) <= height) then return end	-- height check
				
				-- spawn a tracer object
				local box = P_SpawnMobj(found.x, found.y, found.z, MT_KZS_ITEMBOX)
				box.target = ref
				if (found.eflags & MFE_VERTICALFLIP)
					box.eflags = $|(MFE_VERTICALFLIP)
					box.flags2 = $|(MF2_OBJECTFLIP)
				end
				
				-- spawn a special kind of item box that'll always respawn fast
				if (CV_FindVar("respawnitemtime").value)
				and (netgame or multiplayer)
					local newbox = P_SpawnMobj(found.x, found.y, found.z, MT_RANDOMITEM)
					newbox.scale = found.scale
					newbox.kzs_fuse = 3*TICRATE -- fake fuse
					newbox.kzs_fusetarget = box
				end
				P_KillMobj(found)
			end, mo, mo.x - range, mo.x + range, mo.y - range, mo.y + range)
			
			p.kzs.attraction = $+1
			if (p.kzs.attraction == TICRATE)
				kartzshields.RemoveShield(p)
				return
			end
		end,
		
		odds = {0, 2, 2, 1, 0, 0, 0, 0},
	},
	
	-- elemental shield
	{
		name = "Elemental Shield",
		state = S_ELEM1,
		overlays = {S_ELEMF1},
		patch = "K_SHLD_4",
		power = true,
		
		passive = function(p)
			p.kartstuff[k_offroad] = 0	-- lol
		end,
		activate = function(p)
			local mo = p.mo
			
			p.kzs.elemental = 0
			K_DoSneaker(p)
			p.kartstuff[k_sneakertimer] = 8
			K_PlayBoostTaunt(mo)
		end,
		active = function(p)
			local mo = p.mo
			if (p.kzs.elemental % 2)
				-- i REALLY wish i didnt have to remake this
				local ground = mo.floorz
				if (mo.eflags & MFE_VERTICALFLIP)
					ground = mo.ceilingz - FixedMul(mobjinfo[MT_SPINFIRE].height, mo.scale)
				end

				local travelangle = R_PointToAngle2(0, 0, p.rmomx, p.rmomy)
				for i = 0, 1
					local newx = mo.x + P_ReturnThrustX(mo, travelangle + ((i&1) and -1 or 1) * ANGLE_135, FixedMul(FRACUNIT*24, mo.scale))
					local newy = mo.y + P_ReturnThrustY(mo, travelangle + ((i&1) and -1 or 1) * ANGLE_135, FixedMul(FRACUNIT*24, mo.scale))
					if (mo.standingslope)
						ground = P_GetZAt(mo.standingslope, newx, newy)
						if (mo.eflags & MFE_VERTICALFLIP)
							ground = $ - FixedMul(mobjinfo[MT_SPINFIRE].height, mo.scale)
						end
					end
					
					local flame = P_SpawnMobj(newx, newy, ground, MT_KZS_ELEMENTALFIRE)
					flame.color = mo.player.skincolor
					flame.target = mo
					flame.angle = travelangle
					flame.scale = mo.scale
					flame.destscale = mo.destscale
					flame.eflags = (flame.eflags & ~MFE_VERTICALFLIP)|(mo.eflags & MFE_VERTICALFLIP)
				end
			end
			p.kartstuff[k_sneakertimer] = max(2, $)
			
			p.kzs.elemental = $+1
			if (p.kzs.elemental == TICRATE*4/5)
				kartzshields.RemoveShield(p)
				return
			end
		end,
		
		odds = {0, 0, 1, 2, 3, 2, 2, 0},
	},
	
	-- whirlwind shield
	{
		name = "Whirlwind Shield",
		state = S_WIND1,
		patch = "K_SHLD_5",
		power = true,
		
		activate = function(p)
			local mo = p.mo
			
			K_PlayBoostTaunt(mo)
			K_DoPogoSpring(mo, 22<<FRACBITS, 0)
			p.kartstuff[k_pogospring] = 1
			S_StartSound(mo, sfx_wdjump)
		end,
		active = function(p)
			local mo = p.mo
			
			-- end conditions
			if (P_IsObjectOnGround(mo))	-- we're landed
			or not (p.kartstuff[k_pogospring])	-- not springing
				kartzshields.RemoveShield(p)
				return
			end
				
			-- lower gravity
			local modifier = FRACUNIT/2
			mo.momz = $ - (FixedMul(P_GetMobjGravity(mo), modifier))
		end,
		
		odds = {0, 0, 0, 1, 2, 3, 4, 6},
	},
	
	-- armageddon shield
	{
		name = "Armageddon Shield",
		state = S_ARMA1,
		overlays = {S_ARMF1, S_ARMB1},
		patch = "K_SHLD_6",
		power = true,
		
		activate = function(p)
			local mo = p.mo
			local range = FixedMul(mo.scale, FRACUNIT*1280)
			
			-- flashpal players in range!
			for player in players.iterate
				if not (player.mo and player.mo.valid) then continue end
				if (R_PointToDist2(mo.x, mo.y, player.mo.x, player.mo.y) > range) then continue end	-- they're safe...
				P_FlashPal(player, PAL_NUKE, 10)
			end
			
			-- hitfeed support
			local hitfeedHack = false
			if (hitfeed)
				-- this hack is nasty and i fucking hate it
				if not (p.kartstuff[k_sneakertimer])
					p.kartstuff[k_sneakertimer] = 1	-- force a hitfeed message
					hitfeedHack = true
				end
				mo.hf_overrideicon = "K_HMARMA"
			end
			
			-- nuke them!
			P_NukeEnemies(mo, mo, range)
			K_PlayAttackTaunt(mo)
			S_StartSound(mo, sfx_bkpoof)
			kartzshields.RemoveShield(p)
			
			-- hitfeed support (2)
			if (hitfeed)
				if (hitfeedHack) then p.kartstuff[k_sneakertimer] = 0 end
				mo.hf_overrideicon = "K_HMARMA"
			end
		end,
		
		playerspin = function(p)
			kartzshields.Shields[p.kzs.shield].activate(p)
		end,
		playersquish = function(p)
			kartzshields.Shields[p.kzs.shield].activate(p)
		end,
		playerexplode = function(p)
			kartzshields.Shields[p.kzs.shield].activate(p)
		end,
		mobjdamage = function(mo)
			local p = mo.player
			kartzshields.Shields[p.kzs.shield].activate(p)
		end,
		
		odds = {0, 0, 1, 2, 3, 2, 2, 1},
	}
}

kartzshields.AddShield = function(info)
	-- add info to the shields table
	local shieldInfo = {
		-- core stuff
		name = info.name,	-- for disabling it (eventually)
		state = info.state,	-- the state it'll spawn into
		overlays = info.overlays,	-- the overlays that'll spawn on it
		health = info.health,	-- how many hits you can tank
		patch = info.patch,	-- the patch to use for the HUD
		power = info.power, -- is this a power shield? (reduced chance on mash)
		
		-- functions
		get = info.get,
		passive = info.passive,	-- function that'll always be running
		activate = info.activate,	-- function that'll run when activated
		active = info.active,	-- function that'll run after activating the shield
		
		playerspin = info.playerspin,	-- runs in PlayerSpin
		playersquish = info.playersquish,	-- runs in PlayerSquish
		playerexplode = info.playerexplode,	-- runs in PlayerExplode
		mobjdamage = info.mobjdamage,	-- runs in MobjDamage
		
		shouldspin = info.shouldspin,	-- runs in ShouldSpin
		shouldsquish = info.shouldsquish,	-- runs in ShouldSquish
		shouldexplode = info.shouldexplode,	-- runs in ShouldExplode
		shoulddamage = info.shoulddamage,	-- runs in ShouldDamage
	}
	table.insert(kartzshields.Shields, shieldInfo)
	
	-- add to the item odds table
	table.insert(kartzshields.ShieldOdds, info.odds)
	
	-- create a constant for convenience
	local constant = info.name:upper()
	constant = $:gsub(" ", "")
	rawset(_G, "KZS_"..constant, #kartzshields.Shields)
end
for _, info in ipairs(kzs_shields)
	kartzshields.AddShield(info)
end

-- functions to give and remove a shield
kartzshields.SpawnShieldObject = function(p)
	if not (p.kzs and p.kzs.shield) then return end
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	local orb = P_SpawnMobj(mo.x, mo.y, mo.z, MT_KZS_SHIELD)
	orb.target = mo
	if (kartzshields.Shields[p.kzs.shield].state)
		orb.state = kartzshields.Shields[p.kzs.shield].state
	end
	if (kartzshields.Shields[p.kzs.shield].overlays)
		-- spawn overlays
		for _, state in ipairs(kartzshields.Shields[p.kzs.shield].overlays)
			local overlay = P_SpawnMobj(orb.x, orb.y, orb.z, MT_OVERLAY)
			overlay.state = state
			overlay.target = orb
		end
	end
	orb.kzs_shield = p.kzs.shield	-- for validation reasons
	S_StartSound(mo, sfx_shield)
	
	-- store this in the player
	mo.kzs_orb = orb
end

kartzshields.GiveShield = function(p, shield)
	p.kzs.shield = shield
	p.kzs.usingshield = nil
end
kartzshields.RemoveShield = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	p.kzs.shield = nil
	p.kzs.activated = nil
	if (mo.kzs_orb and mo.kzs_orb.valid)
		P_RemoveMobj(mo.kzs_orb)
	end
	mo.kzs_orb = nil
end

if (kartzshields.debug)
	COM_AddCommand("giveshield", function(p, node, shield)
		node = tonumber($)
		local tp = players[node]
		if not (tp and tp.valid) then return end
		
		shield = tonumber($)
		if not (shield)
		or not (kartzshields.Shields[shield])
			return
		end
		
		-- validate if we can set new odds
		kartzshields.GiveShield(tp, shield)
	end, COM_ADMIN)
end
		
-- main shield handler
kartzshields.ShieldThinker = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- lower a player's spinout
	if (mo.kzs_shieldhack)
		if (p.kartstuff[k_spinouttimer])
			p.kartstuff[k_spinouttimer] = max(2, $*7/10)
		end
		if (p.kartstuff[k_squishedtimer])
			p.kartstuff[k_squishedtimer] = max(2, $*7/10)
		end
		
		mo.kzs_shieldhack = false
	end
	
	-- this isn't set? just run this function to remove residual stuff
	if not (p.kzs.shield) then kartzshields.RemoveShield(p) return end
	if not (p.kzs.usingshield)
		-- start using the shield?
		if (p.kzs_shieldbutton == 1)
		and not (p.kartstuff[k_spinouttimer] or p.kartstuff[k_squishedtimer])
			p.kzs.usingshield = true
			
			p.kzs.activated = nil
			p.kzs.health = (kartzshields.Shields[p.kzs.shield].health) or 1
			kartzshields.SpawnShieldObject(p)
			
			-- get function (if it exists)
			if (kartzshields.Shields[p.kzs.shield].get)
				kartzshields.Shields[p.kzs.shield].get(p)
			end
		end
		
		return
	end
	
	-- validate shield object
	if not (mo.kzs_orb and mo.kzs_orb.valid)
		-- spawn it in
		kartzshields.SpawnShieldObject(p)
	end
	
	-- passive handling
	if (kartzshields.Shields[p.kzs.shield].passive)
		kartzshields.Shields[p.kzs.shield].passive(p)
	end
	-- ability handling
	if not (p.kzs.activated)
		if (p.kzs_shieldbutton == 1)
		and (kartzshields.Shields[p.kzs.shield].activate)
		and not (p.kartstuff[k_spinouttimer] or p.kartstuff[k_squishedtimer])
			kartzshields.Shields[p.kzs.shield].activate(p)
			p.kzs.activated = true
		end
	else
		if (kartzshields.Shields[p.kzs.shield].active)
			kartzshields.Shields[p.kzs.shield].active(p)
		end
	end
	
	-- handle shield object
	if (mo.kzs_orb and mo.kzs_orb.valid)	-- paranoia check
		P_MoveOrigin(mo.kzs_orb, mo.x, mo.y, mo.z)
		mo.kzs_orb.scale = mo.scale
		mo.kzs_orb.destscale = mo.destscale
			
		-- drawing conditions
		if (mo.flags2 & MF2_DONTDRAW)
			mo.kzs_orb.flags2 = $|MF2_DONTDRAW
		elseif (p.kzs.activated)
			if not (leveltime%2)
				mo.kzs_orb.flags2 = $|MF2_DONTDRAW
			else
				mo.kzs_orb.flags2 = $ & ~(MF2_DONTDRAW)
			end
		else
			mo.kzs_orb.flags2 = $ & ~(MF2_DONTDRAW)
		end
	end
end